#version 150
precision highp float; precision highp int;
//Ethan Alexander Shulman 2023 "Along the Planes" March 9, 2023

in vec3 in_Position;
out vec4 uv;
out float time;

layout(std140) uniform ShaderviewBlock {
	vec4 mouse;
	float screenX,screenY;
	int frame,framerate;
};
layout(std140) uniform MidiBlock {
	vec4 midi[64];//x = note id, y = velocity, z = channel/track, w = time in seconds since note pressed
	int midiCount;
};
out vec4 song[3];
/*LMMS midi converter https://www.lynxwave.com/LMMStoMIDI/LMMStoMIDIConverter.html since LMMS midi export is broken.

Launch options for midi config 'Shaderview console AlongThePlanes.mid 30 0.75'
30 is the bitmask 0b11110 signifying which instruments are beats/drums with a fixed length.
0.75 is the fixed beat length in seconds.

Midi Instrument 0 = DPiano-A
Midi Instrument 1 = Clap 04 FX.ogg
Midi Instrument 2 = Hithat 02.ogg
Midi Instrument 3 = Kick 07 Euro.ogg
Midi Instrument 4 = Kick 01.ogg*/



#define AA_SAMPLES 16
#define ANIMATION 1
#define EPSILON .01
#define LIGHTING_EPSILON .0001

#define PI 3.141592653589793


mat2 r2d(float a) {
	float c = cos(a), s = sin(a);
	return mat2(c,s,-s,c);
}
//r2 hash http://extremelearning.com.au/unreasonable-effectiveness-of-quasirandom-sequences/
vec2 quasiRandom(float n) {
	#define R2V 1.32471795724474602596
	return fract(.5+n*(1./vec2(R2V,R2V*R2V)));
}


vec2 geometry(vec3 p) {
	//main base cube
	vec2 geo = vec2(length(max(abs(p)-4.5,0.)), 0);
	
	//flip to nearest axes, duplicating song cubes across base cube faces
	vec3 ap = abs(p);
	if (ap.x > ap.y) {
		if (ap.x > ap.z) p = p.yxz;
		else p = p.xzy;
	} else if (p.z > ap.y) p = p.xzy;
	
	//song cubes
	float cd = 1e8;
	for (int y = 0; y < 3; y++) {
		vec4 sv = song[y];
		for (int x = 0; x < 3; x++) {
			float h = sv[x];
			cd = min(cd, length(max(abs(p-vec3(x-1,0,y-1)*2.-vec3(0,4.7+h,0))-vec3(.5,.1+h,.5),0.)));
		}
	}
	if (cd < geo.x) {
		geo = vec2(cd, 1);
	}
	
	return geo;
}
#define geometryDetail geometry

vec3 normal(vec3 p) {
	vec3 n = normalize(vec3(geometryDetail(p+vec3(LIGHTING_EPSILON,0,0)).x,geometryDetail(p+vec3(0,LIGHTING_EPSILON,0)).x,geometryDetail(p+vec3(0,0,LIGHTING_EPSILON)).x)-geometryDetail(p).x);
	if (n != n) return vec3(0,1,0);
	return n;
}


//returns hit position(xyz), material id(w), -1 if no hit
vec4 rayTrace(vec3 rp, vec3 rd, const int maxSteps) {
	for (int i = 0; i < maxSteps && length(rp) < 11.; i++) {
		vec2 geo = geometry(rp);
		if (geo.x < EPSILON) return vec4(rp,geo.y);
		rp += rd*geo.x;
	}
	return vec4(-1);
}


float cubeLighting(vec3 nrm) {
	return dot(normalize(vec3(1,0,0)), nrm)*.5+.5;
}


uniform sampler2D Buffer0;

void vert() {
	#ifdef ANIMATION
	time = float(frame)/float(framerate*AA_SAMPLES);
	
	//initialize song cubes with 0
	for (int i = 0; i < 3; i++) song[i] = vec4(0);
	
	for (int i = 0; i < midiCount; i++) {
		float v = midi[i].z;
		
		if (v == 0.) {//piano
			int n = int(midi[i].x*7.)%9;
			song[n/3][n%3] += .5;
		
		} else if (v == 1.) {//clap
			vec4 cv = vec2(.5*max(0.,1.-midi[i].w/.5),0).xyxy;
			song[0] += cv;
			song[2] += cv;
		
		} else if (v == 2.) {//hat
			float hv = .3*max(0.,1.-midi[i].w/.15);
			song[0].y += hv;
			song[1] += vec4(hv,0,hv,0);
			song[2].y += hv;
			
		} else if (v == 3.) {//kick
			song[1].y += 2.*max(0.,1.-midi[i].w/.5);
			
		}
	}
	
	#else
	song[0] = vec4(0,0,0,0);
	song[1] = vec4(0,1,0,0);
	song[2] = vec4(0,0,0,0);
	time = 0.;
	#endif
	
	uv = vec4((in_Position.xy*vec2(screenX/screenY,1)+(quasiRandom(float(frame%AA_SAMPLES))-.5)*2./screenY),
		(in_Position.xy*.5+.5)*vec2(screenX,screenY));
	gl_Position = vec4(in_Position,1);
}
void frag() {
	vec2 cam = vec2(PI*.25,PI*1.2);
	vec3 rp = vec3(0,0,-10),
		rd = normalize(vec3(uv.xy,1));
	mat2 yaw = r2d(cam.x),
		pitch = r2d(cam.y);
	rp.yz *= pitch;
	rd.yz *= pitch;
	rp.xz *= yaw;
	rd.xz *= yaw;
		

	//camera ray trace
	vec4 hit = rayTrace(rp,rd,80);
	vec3 c;
	if (hit.w < 0.) {
		//background
		c = vec3(.75);
	
	} else {
		//hit, calculate lighting
		vec3 nrm = normal(hit.xyz);
		if (hit.w == 0.) {
			//base cube material
			c = vec3(.8);
			rd = reflect(rd,nrm);
			
			vec4 rhit = rayTrace(hit.xyz+nrm*EPSILON,rd,40);
			if (rhit.w == -1.) c *= .75;
			else c *= .4*cubeLighting(normal(rhit.xyz));
			
		} else {
			c = vec3(.4)*cubeLighting(normal(hit.xyz));	
		}
		
	}
	
	
	
	vec4 sum = texelFetch(Buffer0,ivec2(uv.zw),0);
	if (frame%AA_SAMPLES == 0) sum = vec4(0);
	sum += vec4(c, 1.);
	gl_FragColor = sum;
}
